# -*- coding: utf-8 -*-

import os
from flask import Flask, request, jsonify, g, url_for
from flask_restful import reqparse, abort, Api, Resource, fields, marshal
from flask_sqlalchemy import SQLAlchemy
from flask_httpauth import HTTPBasicAuth
from flask_cors import CORS
from passlib.apps import custom_app_context as pwd_context
from itsdangerous import (TimedJSONWebSignatureSerializer
                          as Serializer, BadSignature, SignatureExpired)
from .common import printWellcome, bcolors, loadConfig, loadMode, resolveRelativeWorkingDirectory, \
    checkDefaultAdminSecret, setAccessLog
import logging
import re
from colorama import init, deinit
import atexit


if 'PYCHARM_HOSTED' not in os.environ:  # Exclude PyCharm IDE from colorama init
    init()
    atexit.register(deinit)


# initialization
app = Flask(__name__)

if loadMode() == True:  # If true debug  mode
    logging.basicConfig(level=logging.DEBUG)
    logger = logging.getLogger(__name__)
else:
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)


config = loadConfig(logger)
try:
    app.config['SECRET_KEY'] = config["app"]["secretKey"]
    app.config['SQLALCHEMY_DATABASE_URI'] = resolveRelativeWorkingDirectory(config["app"]["databaseUri"])
    app.config['SQLALCHEMY_COMMIT_ON_TEARDOWN'] = config["app"]["commitOnTeardown"]
    tokenLife = config["tokenLife"]  # In seconds
    adminSecret = config["adminSecret"]
    checkDefaultAdminSecret(adminSecret)
    cors = config["CORS"]["enable"]
    origins = config["CORS"]["origins"]
    debug = config["debug"]
    setAccessLog(config["app"]["accessLog"])
except KeyError:
    print(bcolors.WARNING+"Misformed config.json!"+bcolors.ENDC)
    exit()

# extensions
db = SQLAlchemy(app)
auth = HTTPBasicAuth()
api = Api(app)
if cors:
    CORS(app, origins=origins)


class User(db.Model):
    __tablename__ = 'users'
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(32), index=True)
    password_hash = db.Column(db.String(64))

    def hash_password(self, password):
        self.password_hash = pwd_context.encrypt(password)

    def verify_password(self, password):
        return pwd_context.verify(password, self.password_hash)

    def generate_auth_token(self, expiration=tokenLife):
        s = Serializer(app.config['SECRET_KEY'], expires_in=expiration)
        return s.dumps({'id': self.id})

    @staticmethod
    def verify_auth_token(token):
        s = Serializer(app.config['SECRET_KEY'])
        try:
            data = s.loads(token)
        except SignatureExpired:
            return None  # valid token, but expired
        except BadSignature:
            return None  # invalid token
        user = User.query.get(data['id'])
        return user


@auth.verify_password
def verify_password(username_or_token, password):
    # first try to authenticate by token
    user = User.verify_auth_token(username_or_token)
    if not user:
        # try to authenticate with username/password
        user = User.query.filter_by(username=username_or_token).first()
        if not user or not user.verify_password(password):
            return False
    g.user = user
    return True


@app.route('/api/users', methods=['POST'])
def new_user():
    if pwd_context.verify(request.values.get('secret'), adminSecret) == False:
        abort(401)    # unauthorized
    username = request.values.get('username')
    password = request.values.get('password')
    if username is None or password is None:
        abort(400)    # missing arguments
    if User.query.filter_by(username=username).first() is not None:
        abort(409)    # existing user
    user = User(username=username)
    user.hash_password(password)
    db.session.add(user)
    db.session.commit()
    return (jsonify({'username': user.username}), 201,
            {'Location': url_for('get_user', id=user.id, _external=True)})


@app.route('/api/users/<int:id>')
@auth.login_required
def get_user(id):
    user = User.query.get(id)
    if not user:
        abort(400)
    return jsonify({'username': user.username})


@app.route('/api/token')
@auth.login_required
def get_auth_token():
    token = g.user.generate_auth_token(tokenLife)
    return jsonify({'token': token.decode('ascii'), 'duration': tokenLife})


@app.route('/api/resource')
@auth.login_required
def get_resource():
    return jsonify({'data': 'Hello, %s!' % g.user.username})


def checkUser(vusers):
    # Check if it is a username or token auth
    user = User.verify_auth_token(auth.username())
    if user:
        return bool(user.username in vusers)
    else:
        return bool(auth.username() in vusers)


def getDomain(request):
    pattern_url_base = re.compile("(http|https):\/\/[a-zA-Z0-9_.]+:[0-9]*")
    return pattern_url_base.search(str(request)).group()


def getResourceURL(endpoint, selector_name = None, selector_value = None, absolute = False):
    if selector_name is None and selector_value is None:
        selector_name = "dummy_RDCH106"
        selector_value = "dummy_RDCH106"
    uri_field = {'url': fields.Url(endpoint)}
    selector = {selector_name: selector_value}
    if absolute:
        return getDomain(request) + marshal(selector, uri_field)["url"]
    else:
        return marshal(selector, uri_field)["url"]


def createDB():
    db_path = str(app.config['SQLALCHEMY_DATABASE_URI'].split('///')[1])
    if not os.path.exists(db_path):
        print(bcolors.WARNING + "Creating SQlite Database" + bcolors.ENDC)
        print(bcolors.OKBLUE + db_path + bcolors.ENDC)
        db.create_all()


def run():
    printWellcome()
    createDB()
    if config["SSL"]["activate"]:
        app.run(host=config["host"]["ip"], port=int(os.environ.get('PORT', config["host"]["port"])), debug=debug,
                processes=config["host"]["processes"], threaded=config["host"]["threaded"],
                ssl_context=(config["SSL"]["certificate"], config["SSL"]["key"]))
    else:
        app.run(host=config["host"]["ip"], port=int(os.environ.get('PORT', config["host"]["port"])), debug=debug,
                processes=config["host"]["processes"], threaded=config["host"]["threaded"])



# -*- coding: utf-8 -*-

import linkero.core.linkero as linkero
from gevent.wsgi import WSGIServer
from gevent.pool import Pool
import os


def run(app):
    linkero.printWellcome()
    linkero.createDB()
    if linkero.config["SSL"]["activate"]:
        gevent_server = WSGIServer((linkero.config["host"]["ip"],
                                    int(os.environ.get('PORT', linkero.config["host"]["port"]))),
                                    app, spawn=Pool(linkero.config["gevent"]["spawn"]),
                                    log='default' if (linkero.config["gevent"]["accessLog"] == True) else None,
                                    keyfile=linkero.config["SSL"]["key"], certfile=linkero.config["SSL"]["certificate"])
    else:
        gevent_server = WSGIServer((linkero.config["host"]["ip"],
                                    int(os.environ.get('PORT', linkero.config["host"]["port"]))),
                                    app, spawn=Pool(linkero.config["gevent"]["spawn"],
                                    log = 'default' if (linkero.config["gevent"]["accessLog"] == True) else None))
    gevent_server.serve_forever()


# -*- coding: utf-8 -*-

import logging
import json
import platform
import os
from shutil import copyfile
from .ascii import ascii_warning


class Metadata:
    def __init__(self):
        self.__version__ = '0.9.9'
        self.__author__ = 'Rubén de Celis Hernández'

    def get_version(self):
        return self.__version__
    def get_author(self):
        return self.__author__


def printWellcome():
    print(bcolors.HEADER+"")
    print("-----------------------------------------------------")
    print(bcolors.OKBLUE+"")
    print("  _____                                "+bcolors.WARNING+"|`-._/\_.-`|"+bcolors.OKBLUE)
    print(" |_   _|                               "+bcolors.WARNING+"|    ||    |"+bcolors.OKBLUE)
    print("   | |  _ __   __ _ _ __ __ _ _ __     "+bcolors.WARNING+"|___o()o___|"+bcolors.OKBLUE)
    print("   | | | '_ \ / _` | '__/ _` | '_ \    "+bcolors.WARNING+"|__((<>))__|"+bcolors.OKBLUE)
    print("  _| |_| | | | (_| | | | (_| | | | |   "+bcolors.WARNING+"\   o\/o   /"+bcolors.OKBLUE)
    print(" |_____|_| |_|\__, |_|  \__,_|_| |_|   "+bcolors.WARNING+" \   ||   / "+bcolors.OKBLUE)
    print("               __/ |                   "+bcolors.WARNING+"  \  ||  /  "+bcolors.OKBLUE)
    print("              |___/                    "+bcolors.WARNING+"   '.||.'   "+bcolors.OKBLUE)
    print(""+bcolors.ENDC)
    print("                                        Engineering  ")
    print(" v"+Metadata().get_version()+bcolors.HEADER+"")
    print("-----------------------------------------------------")
    print(""+bcolors.ENDC)

def loadMode():
    try:
        with open('config/config.json') as config_file:
            try:
                return json.load(config_file)["debug"]
            except KeyError:
                print(bcolors.WARNING+"Misformed config.json!"+bcolors.ENDC)
                exit()
    except IOError:
        print(bcolors.WARNING+"Error loading config.json!"+bcolors.ENDC)
        if not os.path.exists('config'):
            os.makedirs('config')
        copyfile(os.path.dirname(__file__)+'/../config/config.json', 'config/config.json')
        print(bcolors.WARNING+"\nDefault config.json generated! Try again."+bcolors.ENDC)
        exit()

def loadConfig(logger):
    print(bcolors.WARNING)

    if logger.getEffectiveLevel() != logging.DEBUG:
        logger.info("Running in Release Mode")
    else:
        logger.info("Running in Debug Mode")

    try:
        with open('config/config.json') as config_file:
            config = json.load(config_file)
            logger.info("Loaded: config/config.json")
    except IOError:
        print(bcolors.WARNING+"Error loading config.json!"+bcolors.ENDC)
        exit()

    print(bcolors.ENDC)
    return (config)

def setAccessLog(value):
    log = logging.getLogger('werkzeug')
    log.disabled = not value

def resolveRelativeWorkingDirectory(sqlite_path):
    if sqlite_path.find("///", 7) > 0 and sqlite_path.find("////", 7) == -1 and sqlite_path.find(":", 7) == -1:
        return sqlite_path.replace("///", ("///"+os.getcwd()+"/"), 1)

def checkDefaultAdminSecret(adminSecret):
    if adminSecret == "$5$rounds=549561$kWqFvPNTcBsl.Kle$ONikf.BJtqKKFTZbUtVmwZn0nDdwrsHhjyqgxRlUNw4":
        print(bcolors.WARNING + ascii_warning + bcolors.ENDC)
        print(bcolors.FAIL + "You are using 'admin' as admin secret\n" + bcolors.ENDC)
        print("Please generate new one with tools.passwordHashGenerator.generatePasswordHash()")
        print("You can call it typing the following command in your terminal:")
        print(bcolors.OKBLUE + "\n"
              "python -c 'from linkero.tools.passwordHashGenerator import generatePasswordHash; generatePasswordHash()'"
              + "\n" + bcolors.ENDC)

if os.name == 'nt' and platform.release() == '10' and platform.version() >= '10.0.14393':
    # Fix ANSI color in Windows 10 version 10.0.14393 (Windows Anniversary Update)
    import ctypes
    kernel32 = ctypes.windll.kernel32
    kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'



# -*- coding: utf-8 -*-

import sys
import getpass
from linkero.core.common import bcolors
from passlib.apps import custom_app_context as pwd_context

def generatePasswordHash():

    if (sys.version_info > (3, 0)):
        password = input("\nType the password: ")
    else:
        password = raw_input("\nType the password: ")

    password_hash = pwd_context.encrypt(password)

    print("\n"+bcolors.OKBLUE+password_hash+bcolors.ENDC+"\n")

    print(bcolors.WARNING+"Copy the generated password hash in "+bcolors.OKBLUE
        +"adminSecret"+bcolors.WARNING+" param inside of the config.json file"+bcolors.ENDC)






# -*- coding: utf-8 -*-

import linkero.core.linkero as linkero
from waitress import serve
import os


def run(app):
    linkero.printWellcome()
    linkero.createDB()
    serve(app, host=linkero.config["host"]["ip"], port=int(os.environ.get('PORT', linkero.config["host"]["port"])),
          threads=linkero.config["waitress"]["threads"], connection_limit=linkero.config["waitress"]["spawn"])


from setuptools import setup
from linkero.core.common import Metadata

metadata = Metadata()

def requirements():
    """Build the requirements list for this project"""
    requirements_list = []

    with open('requirements.txt') as requirements:
        for install in requirements:
            requirements_list.append(install.strip())

    return requirements_list

long_description = """
Linkero is compatible with **Python-2** and **Python-3**.

************************************
What can I do with Linkero?
************************************

- Run REST API server
- Mount your Python developments or wrappers as API REST service. Only spend time developing your APIs behaviour.
- Add new APIs easily to already existing gateway development.
- Generate Public APIs.
- Generate Private APIs.
    - Security using user and password authentication.
    - Security using token.
    - Grant privileges to desired APIs or only a group of calls of API to specific users.
- User accounts persistence using SQLite DB.
- Integration with other WSGI (Web Server Gateway Interface):
    - Gevent_
    - Waitress_

.. _Gevent: https://github.com/gevent/gevent
.. _Waitress: https://github.com/Pylons/waitress


Installation
============

You can install or upgrade linkero with:

- ``$ pip install linkero --upgrade``

Or you can install from source with:

- ``$ git clone https://github.com/ingran/linkero.git --recursive``

- ``$ cd linkero``

- ``$ pip install .``
   """

setup(
    name = 'linkero',
    packages = ['linkero', 'linkero.core', 'linkero.core.gateway', 'linkero.config', 'linkero.tools'],
    package_data={'linkero': ['config/config.json']},
    install_requires=requirements(),
    version = metadata.get_version(),
    license = 'LGPL v3',
    description = 'Restful API for external requests',
    long_description= long_description,
    author = metadata.get_author(),
    author_email = 'contact@rdch106.hol.es',
    url = 'https://github.com/ingran/linkero',
    download_url = 'https://github.com/ingran/linkero/archive/v'+metadata.get_version()+'.tar.gz',
    keywords = 'framework rest-api restful-api web-services',
    classifiers = ['Programming Language :: Python',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.3',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6'],
)
